/*
    
  This file is a part of Shellp, a 'shell helper' library.
  
  Copyright (C) 2011-2012 Jori Liesenborgs
  
  Contact: jori.liesenborgs@gmail.com

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  
  USA

*/

#include "shellconfig.h"
#include "shellinstance.h"
#include "shellcommand.h"
#include "cmdexit.h"
#include "cmdhelp.h"
#include "cmdeval.h"
#include "cmdecho.h"
#include "cmdexec.h"
#include "cmdimport.h"
#include "cmdexport.h"
#include "cmddefine.h"
#include "cmddir.h"

#include "shellcmdrealarg.h"
#include "shellinstance.h"
#include "iosystem.h"

#include <cmath>

#include "debugnew.h"

using namespace shellp;

std::list<ShellCommand *> registeredCommands;

class CmdTest : public ShellCommand
{
public:
	CmdTest(const std::string &name) : ShellCommand(name)
	{
		m_pArg = new ShellCmdRealArg("x", 0);
		m_pArg->setDescription("argdesc");
		addArgument(m_pArg);

		setDescription("cmddesc");
	}

	~CmdTest()
	{
		delete m_pArg;
	}

	bool execute()
	{
		ShellInstance *pInst = ShellInstance::getInstance();
		IOSystem *pIOSys = pInst->getIOSystem();

		double x = m_pArg->getValue();
		pIOSys->print("%g", x);	
		
		return true;
	}
private:
	ShellCmdRealArg *m_pArg;
};

class CmdPlot2DTest : public ShellCommand
{
public:
	CmdPlot2DTest(const std::string &name) : ShellCommand(name)
	{
		m_pX1 = new ShellCmdRealArg("x1");
		addArgument(m_pX1);

		m_pX2 = new ShellCmdRealArg("x2");
		addArgument(m_pX2);
	}

	~CmdPlot2DTest()
	{
		delete m_pX1;
		delete m_pX2;
	}

	bool execute()
	{
		ShellInstance *pInst = ShellInstance::getInstance();
		IOSystem *pIOSys = pInst->getIOSystem();
		
		std::vector<double> xPoints, yPoints;

		for (double x = m_pX1->getValue() ; x < m_pX2->getValue() ; x += 0.1)
		{
			xPoints.push_back(x);
			yPoints.push_back(std::cos(x));
		}

		if (!pIOSys->plot2D(0, "Cosine", "x", "cos(x)", xPoints, yPoints, true))
		{
			setErrorString("Unable to create 2D plot: " + pIOSys->getErrorString());
			return false;
		}
		return true;
	}
private:
	ShellCmdRealArg *m_pX1;
	ShellCmdRealArg *m_pX2;
};

class CmdPlot3DTest : public ShellCommand
{
public:
	CmdPlot3DTest(const std::string &name) : ShellCommand(name)
	{
	}

	~CmdPlot3DTest()
	{
	}

	bool execute()
	{
		ShellInstance *pInst = ShellInstance::getInstance();
		IOSystem *pIOSys = pInst->getIOSystem();
		
		int numX = 32;
		int numY = 64;
		double xMin = -5;
		double xMax = 5;
		double yMin = -5;
		double yMax = 5;

		std::vector<double> data(numX*numY);

		for (int y = 0 ; y < numY ; y++)
		{
			double yFrac = (double)y/(double)(numY-1);
			double yPos = yFrac * (yMax-yMin) + yMin;

			for (int x = 0 ; x < numX ; x++)
			{
				double xFrac = (double)x/(double)(numX-1);
				double xPos = xFrac * (xMax-xMin) + xMin;

				double val = std::sin(xPos*xPos+yPos*yPos)/(xPos*xPos+yPos*yPos);

				data[x+y*numX] = val;
			}
		}


		if (!pIOSys->plot3D(0, "sin(x^2+y^2)/(x^2+y^2)", "x", "y", "prob", xMin, yMin, (xMax-xMin)/(double)(numX-1),
					(yMax-yMin)/(double)(numY-1), 1, 1, 1, data, numX, numY, true))
		{
			setErrorString("Unable to create 3D plot: " + pIOSys->getErrorString());
			return false;
		}
		return true;
	}
};

void RegisterCommands()
{
	//registeredCommands.push_back(new CmdRestart("restart"));
	registeredCommands.push_back(new CmdExit("exit"));
	registeredCommands.push_back(new CmdExit("quit"));
	registeredCommands.push_back(new CmdHelp("help"));
	registeredCommands.push_back(new CmdExecute("exec"));
	registeredCommands.push_back(new CmdChangeDir("chdir"));
	registeredCommands.push_back(new CmdGetDir("getdir"));
	registeredCommands.push_back(new CmdEcho("echo"));
	registeredCommands.push_back(new CmdImport("import"));
	registeredCommands.push_back(new CmdDefine("def"));
	registeredCommands.push_back(new CmdListDefines("listdef"));
	registeredCommands.push_back(new CmdExport("export"));
	registeredCommands.push_back(new CmdEval("math/def", true, true));
	registeredCommands.push_back(new CmdEval("math/calc", true, false));
	registeredCommands.push_back(new CmdEvalList("math/list"));
	registeredCommands.push_back(new CmdEvalClearVar("math/clearvar"));
	registeredCommands.push_back(new CmdTest("test"));
	registeredCommands.push_back(new CmdPlot2DTest("test2d"));
	registeredCommands.push_back(new CmdPlot3DTest("test3d"));
}

void UnregisterCommands()
{
	while (!registeredCommands.empty())
	{
		delete *(registeredCommands.begin());
		registeredCommands.pop_front();
	}
}

double testFunction(double x)
{
	ShellInstance *pInst = ShellInstance::getInstance();

	return x*pInst->getUnitMultiplier(0);
}

int main(int argc, char **argv)
{
	ShellInstance instance;

	instance.defineConstant("M0", 2e30);
	instance.registerUnitName(0, "Bla");
	instance.setUnitMultiplier(0, 100);
	instance.defineFunction("f", testFunction);

	RegisterCommands();

	int returnStatus = ShellInstance::run("TestShell", "testshell> ", argc, argv);

	UnregisterCommands();

	return returnStatus;
}

