/*
    
  This file is a part of Shellp, a 'shell helper' library.
  
  Copyright (C) 2011-2012 Jori Liesenborgs
  
  Contact: jori.liesenborgs@gmail.com

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  
  USA

*/

#include "cmdhelp.h"
#include "shellcmdstringarg.h"
#include "shellinstance.h"
#include "shellcommand.h"
#include "shellcmdintarg.h"
#include "shellcmdrealarg.h"
#include "shellcmdchoicearg.h"
#include "iosystem.h"
#include <stdio.h>

#include "debugnew.h"

namespace shellp
{

CmdHelp::CmdHelp(const std::string &cmdname) : ShellCommand(cmdname)
{
	m_pHelpCommand = new ShellCmdStringArg("commandname", "*");
	m_pHelpCommand->setDescription("Command name of which help should be retrieved. Use '*' for a list of commands.");
	addArgument(m_pHelpCommand);
	
	setDescription("Request help about a specific command or request a list of commands to be displayed.");
}

CmdHelp::~CmdHelp()
{
	delete m_pHelpCommand;
}

bool CmdHelp::execute()
{
	ShellInstance *pInst = ShellInstance::getInstance();
	IOSystem *m_pIOSys = pInst->getIOSystem();

	if (m_pHelpCommand->getValue() == std::string("*")) // list all commands
	{
		std::vector<std::string> names = pInst->getCommandNames();
		
		m_pIOSys->writeOutputLine("Available commands:");
		for (int i = 0 ; i < names.size() ; i++)
			m_pIOSys->writeOutputLine(std::string("    ") + names[i]);
	}
	else // display help info about a specific command
	{
		ShellCommand *pCmd = pInst->getCommand(m_pHelpCommand->getValue());

		if (pCmd == 0)
		{
			setErrorString(m_pHelpCommand->getValue() + std::string(" is not a registered command name"));
			return false;
		}
		
		std::vector<ShellCmdArg *> args = pCmd->getArguments();
		std::string argStr = ShellCmdArg::getArgumentNames(args);
		
		m_pIOSys->writeOutputLine("");
		m_pIOSys->writeOutputLine("USAGE:");
		m_pIOSys->writeOutputLine(m_pHelpCommand->getValue() + std::string(" ") + argStr);
		m_pIOSys->writeOutputLine("");
		m_pIOSys->writeOutputLine("DESCRIPTION:");
		m_pIOSys->writeOutputLine(pCmd->getDescription());
		m_pIOSys->writeOutputLine("");

		if (args.size() > 0)
		{
			m_pIOSys->writeOutputLine("ARGUMENTS:");	
		
			for (int i = 0 ; i < args.size() ; i++)
			{
				ShellCmdArg *pArg = args[i];
				std::string argInfo;
				char numStr[256];
				
				//snprintf(numStr, 256, "(%d) ", i+1);
				sprintf(numStr, "(%d) ", i+1);

				argInfo = numStr;
				argInfo += pArg->getName();
				argInfo += std::string(" ");
				argInfo += getArgInfo(pArg);
				
				if (pArg->hasDefault())
					argInfo += std::string(" (Default: '") + pArg->getDefault() + std::string("')");

				if (!pArg->shouldExpand())
					argInfo += std::string(" (not expanded)");
				
				m_pIOSys->writeOutputLine(argInfo);
				m_pIOSys->writeOutputLine(pArg->getDescription());

				if (pArg->getArgType() == ShellCmdArg::Choice)
				{
					ShellCmdChoiceArg *pChoiceArg = (ShellCmdChoiceArg *)pArg;
					int numChoices = pChoiceArg->getNumberOfChoices();

					m_pIOSys->writeOutputLine("Valid choices:");
					int maxLen = 0;

					for (int i = 0 ; i < numChoices ; i++)
					{
						std::string name, desc;
						int value;

						pChoiceArg->getChoiceInfo(i, name, desc, value);
						int len = name.length();

						if (len > maxLen)
							maxLen = len;
					}

					for (int i = 0 ; i < numChoices ; i++)
					{
						std::string name, desc;
						std::string spaces;
						int value;

						pChoiceArg->getChoiceInfo(i, name, desc, value);
						int len = name.length();

						spaces.resize(maxLen-len, ' ');
						m_pIOSys->print(" - %s%s : %s", name.c_str(), spaces.c_str(), desc.c_str());
					}
				}

				m_pIOSys->writeOutputLine("");
			}
		}
	}
	return true;
}

std::string CmdHelp::getArgInfo(ShellCmdArg *a) const
{
	ShellCmdArg::ArgType t = a->getArgType();
	std::string str;
	
	if (t == ShellCmdArg::String)
		str += std::string("(String)");
	else if (t == ShellCmdArg::Integer)
		str += std::string("(Integer)");
	else if (t == ShellCmdArg::Real)
	{
		ShellCmdRealArg *pRealArg = (ShellCmdRealArg *)a;

		if (!pRealArg->hasUnitMultiplier())
			str += std::string("(Real)");
		else
		{
			str += std::string("(Real - Unit: ");
			str += pRealArg->getUnitName();
			str += std::string(")");
		}
	}
	else if (t == ShellCmdArg::Bool)
		str += std::string("(Bool)");
	else if (t == ShellCmdArg::Choice)
		str += std::string("(Choice)");
	else
		str += std::string("(unknown type)");

	if (t == ShellCmdArg::Integer)
	{
		ShellCmdIntArg *pArg = (ShellCmdIntArg *)a;
		char str2[1024];
		
		str2[0] = 0;
		
		if (pArg->useMin() && pArg->useMax())
			//snprintf(str2, 1024, " (MIN: %d, MAX: %d)", pArg->getMin(), pArg->getMax());
			sprintf(str2, " (MIN: %d, MAX: %d)", pArg->getMin(), pArg->getMax());
		else
		{
			if (pArg->useMin())
				//snprintf(str2, 1024, " (MIN: %d)", pArg->getMin());
				sprintf(str2, " (MIN: %d)", pArg->getMin());
			if (pArg->useMax())
				//snprintf(str2, 1024, " (MAX: %d)", pArg->getMax());
				sprintf(str2, " (MAX: %d)", pArg->getMax());
		}
		str += std::string(str2);
	}
	else if (t == ShellCmdArg::Real)
	{
		ShellCmdRealArg *pArg = (ShellCmdRealArg *)a;
		char str2[1024];
		
		str2[0] = 0;
		
		if (pArg->useMin() && pArg->useMax())
			//snprintf(str2, 1024, " (MIN: %g, MAX: %g)", (double)pArg->getMin(), (double)pArg->getMax());
			sprintf(str2, " (MIN: %g, MAX: %g)", (double)pArg->getMin(), (double)pArg->getMax());
		else
		{
			if (pArg->useMin())
				//snprintf(str2, 1024, " (MIN: %g)", (double)pArg->getMin());
				sprintf(str2, " (MIN: %g)", (double)pArg->getMin());
			if (pArg->useMax())
				//snprintf(str2, 1024, " (MAX: %g)", (double)pArg->getMax());
				sprintf(str2, " (MAX: %g)", (double)pArg->getMax());
		}
		str += std::string(str2);
	}
	
	return str;
}

} // end namespace
